#include "stdint.h"
#include "stdbool.h"
#include "nRF52840_E0154A07.h"
	
#define DEEP_SLEEP_MODE 0x10
#define DSM_NORMAL 0x00
	
#define SW_RESET 0x12
	
#define SET_ANALOG_BLOCK 0x74
#define AB_DEFAULT 0x54

#define SET_DIGITAL_BLOCK 0x7E
#define DB_DEFAULT 0x3b
	
#define DRIVER_OUTPUT_CONTROL 0x01
	
#define DATA_ENTRY_MODE 0x11
#define DEM_YDEC_XDEC 0x00
#define DEM_YDEC_XINC 0x01
#define DEM_YINC_XDEC 0x02
#define DEM_YINC_XINC 0x03
#define DEM_ADDRESS_Y 0x04
#define DEM_ADDRESS_X 0x00

#define RAM_X_BOUNDS 0x44
#define RAM_Y_BOUNDS 0x45

#define TEMP_SENSOR_CONTROL 0x1A

#define DISPLAY_UPDATE_CONTROL_2 0x22

#define DUC2_LUT_FROM_OTP 0x91
#define DUC2_DISPLAY 0xC7

#define MASTER_ACTIVATION 0x20

#define SET_RAM_X 0x4E
#define SET_RAM_Y 0x4F
	
#define WRITE_RAM_BW 0x24
#define WRITE_RAM_RED 0x26

#define BYTES_PER_LINE 19

static uint8_t bitmapBW[2888];
static uint8_t bitmapR[2888];
extern const uint8_t largeDigits[11][96];
	
static void spiWriteCommand(uint8_t data)
{
	downDC();
	spiWriteByte(data);
}

static void spiWriteData(uint8_t data)
{
	upDC();
	spiWriteByte(data);
}

static void set4wireMode(void)
{
	downBS1();	
}

static void hwResetDisplay(void)
{
	downReset();
	delay_ms(10);
  upReset();
	delay_ms(10);
}

static void swResetDisplay(void)
{
	downCS();
	spiWriteCommand(SW_RESET);
	upCS();
	delay_ms(1);	
	while(isBusyHigh());
	delay_ms(1);
}

static void wakeUpDisplay(void)
{
	downCS();
  spiWriteCommand(DEEP_SLEEP_MODE);
  spiWriteData(DSM_NORMAL);
	upCS();
	delay_ms(1);
}

static void configureDisplay(void)
{
	downCS();
  spiWriteCommand(SET_ANALOG_BLOCK); 
  spiWriteData(AB_DEFAULT);
	upCS();
	delay_ms(1);
	downCS();
  spiWriteCommand(SET_DIGITAL_BLOCK); 
  spiWriteData(DB_DEFAULT);
	upCS();
	delay_ms(1);
	downCS();
  spiWriteCommand(DRIVER_OUTPUT_CONTROL); 
  spiWriteData(0xD3);
  spiWriteData(0x00);
  spiWriteData(0x00);
	upCS();
	delay_ms(1);
	downCS();
	spiWriteCommand(DATA_ENTRY_MODE);
	spiWriteData(DEM_YINC_XINC | DEM_ADDRESS_X);
	upCS();
	delay_ms(1);
	downCS();
  spiWriteCommand(RAM_X_BOUNDS);
  spiWriteData(0x00);
  spiWriteData(0x12);
	upCS();
	delay_ms(1);
	downCS();
  spiWriteCommand(RAM_Y_BOUNDS);
  spiWriteData(0x00);
  spiWriteData(0x00);
  spiWriteData(0x97); 
  spiWriteData(0x00); 
	upCS();
	delay_ms(1);
	downCS();
  spiWriteCommand(TEMP_SENSOR_CONTROL);
  spiWriteData(0x19);
  spiWriteData(0x00);
	upCS();
	delay_ms(1);
}

static void loadLutFromOtp(void)
{
	downCS();
  spiWriteCommand(DISPLAY_UPDATE_CONTROL_2);
  spiWriteData(DUC2_LUT_FROM_OTP);
	upCS();
	delay_ms(1);
	downCS();
	spiWriteCommand(MASTER_ACTIVATION);
	upCS();
	delay_ms(1);
	while(isBusyHigh());
}

static void moveTo00(void)
{
	downCS();
	spiWriteCommand(SET_RAM_X);
	spiWriteData(0x00);
	upCS();
	delay_ms(1);
	downCS();
	spiWriteCommand(SET_RAM_Y);
	spiWriteData(0x00);
	spiWriteData(0x00);
	upCS();
	delay_ms(1);
}

void drawBuffers(void)
{
	moveTo00();
	downCS();
	spiWriteCommand(WRITE_RAM_BW);
	for(uint32_t i=0; i<2888; i++)spiWriteData(bitmapBW[i]);
	upCS();
	
	moveTo00();
	downCS();
	spiWriteCommand(WRITE_RAM_RED);
	for(uint32_t i=0; i<2888; i++)spiWriteData(bitmapR[i]);
	upCS();
	
	downCS();
  spiWriteCommand(DISPLAY_UPDATE_CONTROL_2);
  spiWriteData(DUC2_DISPLAY);
	upCS();
	delay_ms(10);
	downCS();
	spiWriteCommand(MASTER_ACTIVATION);
	upCS();
	delay_ms(10);
	
	while(isBusyHigh());
}

void clearBuffers(void)
{
	for(uint32_t i=0; i<2888; i++) bitmapBW[i] = 0xFF;
	for(uint32_t i=0; i<2888; i++) bitmapR[i] = 0x00;
}

void drawImageToBuffer(const uint8_t * image, bool red)
{
	if (red){
		for(uint32_t i=0; i<2888; i++){
			bitmapR[i] = image[i];
		}
	}
	else
	{
		for(uint32_t i=0; i<2888; i++){
			bitmapBW[i] = (~image[i]);
			bitmapR[i] &= (~image[i]);
		}
	}
}

void drawDigit(uint32_t x, uint32_t y, uint8_t digit, bool red)
{
	uint8_t offset = x % 8;
	uint32_t xByteNumber = x / 8;
	
	uint8_t byteValue = 0x00;
	uint8_t oldValue;

	for(uint32_t i=0;i<4;i++)
	{
		for(uint32_t j=0;j<32;j++)
		{
			if(xByteNumber+i<19)
			{
				if(red)
				{
					oldValue = bitmapR[xByteNumber+i+BYTES_PER_LINE*(y+j)];
					byteValue = 0x00;
					if(i<3)byteValue = ((largeDigits[digit][j*3+i])>>offset);
					if(i>0)byteValue = byteValue | ((largeDigits[digit][j*3+i-1])<<(8-offset));
					bitmapR[xByteNumber+i+BYTES_PER_LINE*(y+j)] = byteValue | oldValue;
				}
				else
				{
					oldValue = bitmapBW[xByteNumber+i+BYTES_PER_LINE*(y+j)];
					byteValue = 0x00;
					if(i<3)byteValue = ((largeDigits[digit][j*3+i])>>offset);
					if(i>0)byteValue = byteValue | ((largeDigits[digit][j*3+i-1])<<(8-offset));
					bitmapR[xByteNumber+i+BYTES_PER_LINE*(y+j)] &= (~byteValue);
					bitmapBW[xByteNumber+i+BYTES_PER_LINE*(y+j)] = oldValue &(~byteValue);
				}
			}
		}
	}
}

void drawLongNumber(uint32_t x, uint32_t y, int32_t number, uint32_t maxDigits, bool red, bool valueNegative)
{
	uint8_t digit[6];
	uint8_t numberOfDigits;
	uint8_t minDigits = 1;
	
	if(number<0)
	{
		number = -number;
	}
	
	for(uint8_t i=0;i<maxDigits;i++)
	{
		digit[i] = number % 10;
		number = number / 10;
		if(digit[i] != 0) numberOfDigits = i+1;
	}
	
	if (numberOfDigits < minDigits) numberOfDigits = minDigits;
	
	for(int32_t i = 0;i<numberOfDigits; i++)
	{
		uint32_t xPosition = x + ((maxDigits-1)*8*3) - i*8*3;
		drawDigit(xPosition,y,digit[i],red);
	}
	if(valueNegative)
	{
		uint32_t xPosition = x + ((maxDigits-1)*8*3) - numberOfDigits*8*3;
		drawDigit(xPosition,y,10,red);
	}
}

void displayOn(void)
{
	upDC();
	upCS();
	downOn();
}

void displayOff(void)
{
	downBS1();
	downDC();
	downCS();
	downReset();
	upOn();
}

void initHink(void)
{
	displayOn();
	set4wireMode();
	hwResetDisplay();
	swResetDisplay();
	wakeUpDisplay();
	configureDisplay();
	loadLutFromOtp();
}

void initHinkHardware(void)
{
	initDisplayHw();
}

