//#include "nrf_log.h"
//#include "nrf_log_ctrl.h"
//#include "app_timer.h"
#include "nrf_ble_gatt.h"
#include "nrf_sdm.h"
#include "softdevice_handler.h"
#include "ble_hci.h"
#include "ble_db_discovery.h"
#include "nrf_nvic.h"

#include "SEGGER_RTT.h"
#include "HINK_E0154A07.h"
#include "nRF52840_E0154A07.h"


#define CENTRAL_LINK_COUNT        1 
#define SCAN_INTERVAL             0x00A0                                /**< Determines scan interval in units of 0.625 millisecond. */
#define SCAN_WINDOW               0x0050                                /**< Determines scan window in units of 0.625 millisecond. */
#define SCAN_TIMEOUT              0x0000                                /**< Timout when scanning. 0x0000 disables timeout. */

#define MIN_CONNECTION_INTERVAL   MSEC_TO_UNITS(7.5, UNIT_1_25_MS)      /**< Determines minimum connection interval in milliseconds. */
#define MAX_CONNECTION_INTERVAL   MSEC_TO_UNITS(30, UNIT_1_25_MS)       /**< Determines maximum connection interval in milliseconds. */
#define SLAVE_LATENCY             0                                     /**< Determines slave latency in terms of connection events. */
#define SUPERVISION_TIMEOUT       MSEC_TO_UNITS(4000, UNIT_10_MS)       /**< Determines supervision time-out in units of 10 milliseconds. */

uint16_t connectionHandle;

uint16_t tempHandle;
uint16_t tempCccdHandle;

uint16_t pressureHandle;
uint16_t pressureCccdHandle;

uint16_t humidHandle;
uint16_t humidCccdHandle;

uint16_t airqtHandle;
uint16_t airqtCccdHandle;

uint16_t configHandle;

#define TEMP_CHAR 0x0201
#define PRESS_CHAR 0x0202
#define HUMID_CHAR 0x0203
#define AIRQT_CHAR 0x0204
#define CONF_CHAR 0x206

#define WEATHER_SERVICE 0x0200

static char const m_target_periph_name[] = "Thingy";             /**< Name of the device we try to connect to. This name is searched in the scan report data*/
static nrf_ble_gatt_t     m_gatt;               /**< GATT module instance. */
static uint8_t weatherUuidType;

extern const uint8_t WS_InterfaceR [];
extern const uint8_t WS_InterfaceBW [];

extern const uint8_t largeDigits[10][96];

static volatile int8_t tempVal;
static volatile int8_t tempValDecimal;

static volatile int32_t pressVal;
static volatile int32_t humidVal;

static volatile bool tempCollected, pressCollected, humidCollected;

typedef struct
{
    uint8_t * p_data;   /**< Pointer to data. */
    uint16_t  data_len; /**< Length of data. */
} data_t;

static ble_db_discovery_t m_ble_db_discovery;   /**< DB structure used by the database discovery module. */
static ble_gap_scan_params_t const m_scan_params =
{
    .active   = 1,
    .interval = SCAN_INTERVAL,
    .window   = SCAN_WINDOW,
    .timeout  = SCAN_TIMEOUT,
    #if (NRF_SD_BLE_API_VERSION <= 2)
        .selective   = 0,
        .p_whitelist = NULL,
    #endif
    #if (NRF_SD_BLE_API_VERSION >= 3)
        .use_whitelist = 0,
    #endif
};

static ble_gap_conn_params_t const m_connection_param =
{
    (uint16_t)MIN_CONNECTION_INTERVAL,
    (uint16_t)MAX_CONNECTION_INTERVAL,
    (uint16_t)SLAVE_LATENCY,
    (uint16_t)SUPERVISION_TIMEOUT
};

static uint8_t txCccdValue[2];
ble_gattc_write_params_t gattcParams;
volatile uint32_t returnValue;


static void cccdConfigure(uint16_t conn_handle, uint16_t handle_cccd, bool enable)
{
	uint16_t cccdValue;
	char val[20];

	
	if(enable)cccdValue = BLE_GATT_HVX_NOTIFICATION;
	else cccdValue = 0;

	txCccdValue[0] = LSB_16(cccdValue);
	txCccdValue[1] = MSB_16(cccdValue);
	
	
	gattcParams.handle = handle_cccd;
	gattcParams.len = BLE_CCCD_VALUE_LEN;
	gattcParams.p_value = txCccdValue;
	gattcParams.offset = 0;
	gattcParams.write_op = BLE_GATT_OP_WRITE_REQ;

	returnValue = sd_ble_gattc_write(conn_handle,&gattcParams);
	sprintf(val,"%d\r\n",returnValue);
}


static void temperatureNotifyEnable(void)
{
	cccdConfigure(connectionHandle,tempCccdHandle,true);
}

static void pressureNotifyEnable(void)
{
	cccdConfigure(connectionHandle,pressureCccdHandle,true);
}

static void humidityNotifyEnable(void)
{
	cccdConfigure(connectionHandle,humidCccdHandle,true);
}

static void airqtNotifyEnable(void)
{
	cccdConfigure(connectionHandle,airqtCccdHandle,true);
}

static void bleOnWeatherEvent(ble_evt_t * p_ble_evt)
{
	char tempString[50];
	int32_t * pressure;
	uint16_t *eCO2, *TVOC;
    if (p_ble_evt == NULL)
    {
        return;
    }
		
    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GATTC_EVT_HVX:
					if(p_ble_evt->evt.gattc_evt.params.hvx.handle == tempHandle)
					{
					//	sprintf(tempString,"T = %d.%d\xB0\x43\r\n",p_ble_evt->evt.gattc_evt.params.hvx.data[0],p_ble_evt->evt.gattc_evt.params.hvx.data[1]);
				//		refreshCounter ++;
						tempVal = p_ble_evt->evt.gattc_evt.params.hvx.data[0];
						tempValDecimal = p_ble_evt->evt.gattc_evt.params.hvx.data[1];
						tempCollected = true;

					}
					else
					if(p_ble_evt->evt.gattc_evt.params.hvx.handle == pressureHandle)
					{
						pressure = (int32_t *)&p_ble_evt->evt.gattc_evt.params.hvx.data[0];
						pressVal = *pressure;
				//		sprintf(tempString,"P = %d.%d hPa\r\n",*pressure,p_ble_evt->evt.gattc_evt.params.hvx.data[4]);
						pressCollected = true;
					}
					else
					if(p_ble_evt->evt.gattc_evt.params.hvx.handle == humidHandle)
					{
					//	sprintf(tempString,"RH = %d%%\r\n",p_ble_evt->evt.gattc_evt.params.hvx.data[0]);
						humidVal = p_ble_evt->evt.gattc_evt.params.hvx.data[0];
						humidCollected = true;
					}
					else
					if(p_ble_evt->evt.gattc_evt.params.hvx.handle == airqtHandle)
					{
						eCO2 = (uint16_t *)&p_ble_evt->evt.gattc_evt.params.hvx.data[0];
						TVOC = (uint16_t *)&p_ble_evt->evt.gattc_evt.params.hvx.data[2];
						sprintf(tempString,"CO2 = %d ppm\r\nTVOC = %d ppb\r\n",*eCO2,*TVOC);
					}

//					SEGGER_RTT_WriteString(0,tempString);
          break;

        case BLE_GATTC_EVT_WRITE_RSP:
//					SEGGER_RTT_WriteString(0,"rsp\r\n");
					if(p_ble_evt->evt.gattc_evt.params.write_rsp.handle ==  tempCccdHandle)
					{
						pressureNotifyEnable();
					}
					else if(p_ble_evt->evt.gattc_evt.params.write_rsp.handle ==  pressureCccdHandle)
					{
						humidityNotifyEnable();
					}
					else if(p_ble_evt->evt.gattc_evt.params.write_rsp.handle ==  humidCccdHandle)
					{
						airqtNotifyEnable();
					}
					else if(p_ble_evt->evt.gattc_evt.params.write_rsp.handle ==  airqtCccdHandle)
					{
					}
        break;

        case BLE_GAP_EVT_DISCONNECTED:
//            on_disconnected(p_ble_lbs_c, p_ble_evt);
            break;

        default:
            break;
    }
}


static uint32_t adv_report_parse(uint8_t type, data_t * p_advdata, data_t * p_typedata)
{
    uint32_t  index = 0;
    uint8_t * p_data;

    p_data = p_advdata->p_data;

    while (index < p_advdata->data_len)
    {
        uint8_t field_length = p_data[index];
        uint8_t field_type   = p_data[index + 1];

        if (field_type == type)
        {
            p_typedata->p_data   = &p_data[index + 2];
            p_typedata->data_len = field_length - 1;
            return NRF_SUCCESS;
        }
        index += field_length + 1;
    }
    return NRF_ERROR_NOT_FOUND;
}

static void on_adv_report(const ble_evt_t * const p_ble_evt)
{
    ret_code_t err_code;
    data_t     adv_data;
    data_t     dev_name;
    bool       do_connect = false;

    ble_gap_evt_t  const * p_gap_evt  = &p_ble_evt->evt.gap_evt;
    ble_gap_addr_t const * peer_addr  = &p_gap_evt->params.adv_report.peer_addr;

    adv_data.p_data   = (uint8_t *)p_gap_evt->params.adv_report.data;
    adv_data.data_len = p_gap_evt->params.adv_report.dlen;

    bool name_found = false;
    err_code = adv_report_parse(BLE_GAP_AD_TYPE_COMPLETE_LOCAL_NAME, &adv_data, &dev_name);

    if (err_code != NRF_SUCCESS)
    {
        err_code = adv_report_parse(BLE_GAP_AD_TYPE_SHORT_LOCAL_NAME, &adv_data, &dev_name);
        if (err_code != NRF_SUCCESS)
        {
            return;
        }
        else
        {
            name_found = true;
        }
    }
    else
    {
        name_found = true;
    }

    if (name_found)
    {
        if (strlen(m_target_periph_name) != 0)
        {
            if (memcmp(m_target_periph_name, dev_name.p_data, dev_name.data_len )== 0)
            {
                do_connect = true;
            }
        }
    }

    if (do_connect)
    {
        err_code = sd_ble_gap_connect(peer_addr,
                                      &m_scan_params,
                                      &m_connection_param,
                                      BLE_CONN_CFG_TAG_DEFAULT);
        APP_ERROR_CHECK(err_code);
    }
}


static void scan_start(void)
{
    ret_code_t err_code;

    (void) sd_ble_gap_scan_stop();

    err_code = sd_ble_gap_scan_start(&m_scan_params);
    APP_ERROR_CHECK(err_code);
}

static void on_ble_evt(ble_evt_t const * p_ble_evt)
{
    ret_code_t err_code;

    // For readability.
    ble_gap_evt_t const * p_gap_evt = &p_ble_evt->evt.gap_evt;

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
        {
            err_code = ble_db_discovery_start(&m_ble_db_discovery, p_gap_evt->conn_handle);
            APP_ERROR_CHECK(err_code);
        } break;

        case BLE_GAP_EVT_DISCONNECTED:
        {
            scan_start();
        } break;

        case BLE_GAP_EVT_ADV_REPORT:
        {
            on_adv_report(p_ble_evt);
        } break;

        case BLE_GAP_EVT_TIMEOUT:
        {
            if (p_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_CONN)
            {
            }
        } break;

        case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST:
        {
            err_code = sd_ble_gap_conn_param_update(p_gap_evt->conn_handle,
                                        &p_gap_evt->params.conn_param_update_request.conn_params);
            APP_ERROR_CHECK(err_code);
        } break;

        case BLE_GATTC_EVT_TIMEOUT:
        {
            err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            APP_ERROR_CHECK(err_code);
        } break;

        case BLE_GATTS_EVT_TIMEOUT:
        {
            err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            APP_ERROR_CHECK(err_code);
        } break;

        default:
            break;
    }
}


static void ble_evt_dispatch(ble_evt_t * p_ble_evt)
{
  on_ble_evt(p_ble_evt);
  ble_db_discovery_on_ble_evt(&m_ble_db_discovery, p_ble_evt);
	bleOnWeatherEvent(p_ble_evt);
  nrf_ble_gatt_on_ble_evt(&m_gatt, p_ble_evt);
}


static void ble_stack_init(void)
{
    ret_code_t err_code;

    nrf_clock_lf_cfg_t clock_lf_cfg = {.source        = NRF_CLOCK_LF_SRC_XTAL,           
                                 .rc_ctiv       = 0,                                
                                 .rc_temp_ctiv  = 0,                                
                                 .xtal_accuracy = NRF_CLOCK_LF_XTAL_ACCURACY_20_PPM};
    SOFTDEVICE_HANDLER_INIT(&clock_lf_cfg, NULL);

    uint32_t ram_start = 0;
    err_code = softdevice_app_ram_start_get(&ram_start);
    APP_ERROR_CHECK(err_code);

    ble_cfg_t ble_cfg;

    memset(&ble_cfg, 0, sizeof(ble_cfg));
    ble_cfg.gap_cfg.role_count_cfg.periph_role_count  = 0;
    ble_cfg.gap_cfg.role_count_cfg.central_role_count = CENTRAL_LINK_COUNT;
    ble_cfg.gap_cfg.role_count_cfg.central_sec_count  = 1;
    err_code = sd_ble_cfg_set(BLE_GAP_CFG_ROLE_COUNT, &ble_cfg, ram_start);
    APP_ERROR_CHECK(err_code);

    memset(&ble_cfg, 0, sizeof(ble_cfg));
    ble_cfg.gatts_cfg.attr_tab_size.attr_tab_size = BLE_GATTS_ATTR_TAB_SIZE_MIN;
    err_code = sd_ble_cfg_set(BLE_GATTS_CFG_ATTR_TAB_SIZE, &ble_cfg, ram_start);
    APP_ERROR_CHECK(err_code);

    err_code = softdevice_enable(&ram_start);
    APP_ERROR_CHECK(err_code);

    err_code = softdevice_ble_evt_handler_set(ble_evt_dispatch);
    APP_ERROR_CHECK(err_code);
}

static void db_disc_handler(ble_db_discovery_evt_t * p_evt)
{
	uint32_t i;
	
	if(
		(p_evt->conn_handle == connectionHandle)
		&&(p_evt->evt_type == BLE_DB_DISCOVERY_COMPLETE)
		&&(p_evt->params.discovered_db.srv_uuid.uuid == WEATHER_SERVICE)
		&&(p_evt->params.discovered_db.srv_uuid.type == weatherUuidType)
	)
	{
		connectionHandle = p_evt->conn_handle;
			
    for (i = 0; i < p_evt->params.discovered_db.char_count+1; i++)
    {
			const ble_gatt_db_char_t * p_char = &(p_evt->params.discovered_db.charateristics[i]);
      switch (p_char->characteristic.uuid.uuid)
      {
				case TEMP_CHAR:
					tempHandle = p_char->characteristic.handle_value; 
					tempCccdHandle = p_char->cccd_handle;
				break;
				case PRESS_CHAR:
					pressureHandle = p_char->characteristic.handle_value; 
					pressureCccdHandle = p_char->cccd_handle;
				break;
				case HUMID_CHAR:
					humidHandle = p_char->characteristic.handle_value; 
					humidCccdHandle = p_char->cccd_handle;
				break;
				case AIRQT_CHAR:
					airqtHandle = p_char->characteristic.handle_value; 
					airqtCccdHandle = p_char->cccd_handle;
				break;
				case CONF_CHAR:
					configHandle = p_char->characteristic.handle_value; 
				break;
        default:break;
      }
    }
		temperatureNotifyEnable();
	}
}

static uint8_t weatherUuidType;

#define THINGY_UUID_BASE        {0x42, 0x00, 0x74, 0xA9, 0xFF, 0x52, 0x10, 0x9B, \
                              0x33, 0x49, 0x35, 0x9B, 0x00, 0x00, 0x68, 0xEF}

#define WEATHER_SERVICE 0x0200


static void weatherServiceInit(void)
{
  ble_uuid_t    weatherUuid;
  ble_uuid128_t thingyBaseUuid = {THINGY_UUID_BASE};

	sd_ble_uuid_vs_add(&thingyBaseUuid, &weatherUuidType);

	weatherUuid.type = weatherUuidType;
	weatherUuid.uuid = WEATHER_SERVICE;
	
	ble_db_discovery_evt_register(&weatherUuid);
}

void led1init(void)
{
	NRF_P0->PIN_CNF[13] = GPIO_PIN_CNF_DIR_Output | GPIO_PIN_CNF_DRIVE_S0S1;
	NRF_P0->OUTSET = 0x2000;
}

void led1on(void)
{
	NRF_P0->OUTCLR = 0x2000;
}

void led1off(void)
{
	NRF_P0->OUTSET = 0x2000;
}

void led2init(void)
{
	NRF_P0->PIN_CNF[14] = GPIO_PIN_CNF_DIR_Output | GPIO_PIN_CNF_DRIVE_S0S1;
	NRF_P0->OUTSET = 0x4000;
}

void led2on(void)
{
	NRF_P0->OUTCLR = 0x4000;
}

void led2off(void)
{
	NRF_P0->OUTSET = 0x4000;
}

void run32kHzClock(void)
{
	NRF_CLOCK->TASKS_LFCLKSTOP = 1;
	delay_ms(1);
	NRF_CLOCK->LFCLKSRC = NRF_CLOCK_LF_SRC_XTAL;
	delay_ms(1);
	NRF_CLOCK->TASKS_LFCLKSTART = 1;
	delay_ms(1);
}

void initRtc(void)
{
	run32kHzClock();
	NRF_RTC1->TASKS_STOP = 1;
	delay_ms(1);
	
	NRF_RTC1->PRESCALER = 0xFFF;
	NRF_RTC1->CC[0] = 8*150;

  NRF_RTC1->TASKS_CLEAR = 1;
	delay_ms(1);
	NRF_RTC1->EVENTS_COMPARE[0] = 0;
	delay_ms(1);
	NRF_RTC1->EVTEN = RTC_EVTEN_COMPARE0_Msk;
	delay_ms(1);
	NRF_RTC1->TASKS_START = 1;
}

void waitForRtc(void)
{
	while(NRF_RTC1->EVENTS_COMPARE[0] == 0);
}

void refreshDisplay(void)
{
	bool negative;

	initHink();	
	clearBuffers();
	
	drawImageToBuffer(WS_InterfaceR,true);
	drawImageToBuffer(WS_InterfaceBW,false);

	if((tempVal<0)||(tempValDecimal<0))negative = true;
	else negative = false;
	drawLongNumber(82,4,(tempValDecimal/10),1,false,false);
	drawLongNumber(27,4,tempVal,2,false,negative);
	drawLongNumber(10,42,pressVal,4,false,false);
	drawLongNumber(34,80,humidVal,3,false,false);
	drawBuffers();
}

int main(void)
{
	initHinkHardware();
	led1init();
	led2init();
	displayOff();
  while(true)
  {
		tempCollected  = false;
		pressCollected = false;
		humidCollected = false;

		ble_stack_init();
		nrf_ble_gatt_init(&m_gatt, NULL);
		ble_db_discovery_init(db_disc_handler);
		weatherServiceInit();
		
		led1on();
		scan_start();

		bool timeOut = false;
		bool allMeasurementsCollected = false;
		while((!allMeasurementsCollected)&&(!timeOut))
		{
			sd_app_evt_wait();
			if(tempCollected && pressCollected && humidCollected) allMeasurementsCollected = true;
		}
		softdevice_handler_sd_disable();
		initRtc();
		led1off();
		if(allMeasurementsCollected)
		{
			led2on();
			refreshDisplay();
			led2off();
		}
		waitForRtc();
  }
}
